# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import awkward1.layout

anycontent = (
  awkward1.layout.NumpyArray,
  awkward1.layout.ListArray32,
  awkward1.layout.ListArray64,
  awkward1.layout.ListOffsetArray32,
  awkward1.layout.ListOffsetArray64,
  )
